"""call1D.py--- module that calls 1D MPPD code and reads results back into sim object"""

import simclass
import numpy as np
import sys
import os
import csv
from data import name1D, shortname1D
from post3D import write_csv_3D1D 
import yaml

TOL=1.e-8
small=lambda arr : np.amax(np.abs(arr)) < TOL

def call1D(sim: simclass.SimClass):
    """Call 1D MPPD code"""

    tstart3D=sim.inhales[sim.currbreath][0]
    tendinh3D=sim.inhales[sim.currbreath][1]
    tendexh3D=sim.exhales[sim.currbreath][1]
    ntimes=round((tendexh3D-tstart3D)/sim.dtio)
# We take airvol_pmass_outlet<num>.csv (which includes output
# computed from tstart3D to tendinh3D)
# and pass it to MPPD.jar for each subtree.  Then we collect output
# from mppdout_exh_<num>.csv, mppdout_dep_<num>.csv and use it write
# sim.mass1Dto3D from tendinh3D to tendexh3D and 
# sim.mass1Ddeposited from tstart3D to tendexh3D and 
# sim.mass1Dsuspended from tstart3D to tendexh3D.

    csvtime=np.zeros(len(sim.time))
    csvexh=np.zeros(len(sim.time))
    csvgen=np.zeros(len(sim.time)*sim.ngen,dtype='i')
    csvdep=np.zeros(len(sim.time)*sim.ngen)
    csvsus=np.zeros(len(sim.time)*sim.ngen)
    writedir=f'multiResults/{sim.currbreath}_inh'
    for idist in range(sim.ndist):
        with open(f'{writedir}/params{shortname1D[idist]}.csv','w') as f:
            f.write(f'{shortname1D[idist]}\n')
            f.write(f'{sim.particleDiameter}\n')
            f.write(f'{sim.particleDensity}\n')
            f.write(f'{tendinh3D-tstart3D}\n')
            f.write(f'{tendexh3D-tendinh3D}\n')
            f.write(f'{sim.dtio}\n')
            f.write(f'{sim.params1D["segmentsPerAirway"]}\n')
            f.write(f'{sim.params1D["MixingModel"]}\n')
            f.write(f'{sim.params1D["FRCinMl"][idist]}\n')

        if sim.currbreath==0:
            # Initial MPPD call without state file (since zero particles exist at beginning)
            com=f'source appenv.sh; java -jar MPPD.jar {writedir}/params{shortname1D[idist]}.csv'\
                f' pnnloutlet{shortname1D[idist]}.dat'\
                f' {writedir}/airvol_pmass_{name1D[idist]}.csv {writedir}/mppdout '
        else:
            # Subsequent MPPD call with state file (since particles exist in subtree)
            statedir=f'multiResults/{sim.currbreath-1}_inh'
            com=f'source appenv.sh; java -jar MPPD.jar {writedir}/params{shortname1D[idist]}.csv'\
                f' pnnloutlet{shortname1D[idist]}.dat'\
                f' {writedir}/airvol_pmass_{name1D[idist]}.csv {writedir}/mppdout '\
                f' {statedir}/mppdout_state_{shortname1D[idist]}.csv'
        print(f' Attempting to execute:\n{com}\n ',flush=True)
        if os.system(com):
            sys.exit(f'call1D: Problem with {name1D[idist]}')
        with open(f'{writedir}/mppdout_exh_{shortname1D[idist]}.csv','r') as f:
            #Open output from MPPD and write subtree exhalation data into sim object
            csv_reader=csv.reader(f)
            for lnum,line in enumerate(csv_reader):
                nhead=2        # Second header line because Owen repeated the 0.0 datum
                if lnum >= nhead:
                    csvtime[lnum-nhead]=float(line[0])
                    csvexh[lnum-nhead]=float(line[1])
                    numfiletimes=lnum-nhead+1
            a,b=samesubset(tstart3D,ntimes+1,sim.time,csvtime+tstart3D)
            for i in range(1,ntimes+1):
                sim.mass1Dto3D[idist,a+i]=csvexh[b+i]+sim.mass1Dto3D[idist,a]
        with open(f'{writedir}/mppdout_dep_{shortname1D[idist]}.csv','r') as f:
            #Open output from MPPD and write subtree deposition data into sim object
            csv_reader=csv.reader(f)
            for lnum,line in enumerate(csv_reader):
                nhead=1
                if lnum >= nhead:
                    csvgen[lnum-nhead]=int(line[0])
                    csvtimedep=float(line[1])
                    csvdep[lnum-nhead]=float(line[2])
# Assume time given in same order as for exhalation.  But test this.
                    modind=(lnum-nhead)%numfiletimes
                    assert small(csvtime[modind]-csvtimedep), "Exh,Dep file time problem"
                    numfiledatalines_dep=lnum-nhead+1
            assert numfiledatalines_dep%numfiletimes==0, "Exh, Dep file numgen problem"
            numgen=round(numfiledatalines_dep / numfiletimes)
            # The data items we want to write are b+k*numfiletimes+m
            # for k=0... numgen-1 and m=1... ntimes 
            for k in range(numgen):
                for m in range(1,ntimes+1):
                    dataitem=b+k*numfiletimes+m
                    sim.mass1Ddeposited[idist,csvgen[dataitem],m+a]=( csvdep[dataitem]+
                            sim.mass1Ddeposited[idist,csvgen[dataitem],a] )
        with open(f'{writedir}/mppdout_sus_{shortname1D[idist]}.csv','r') as f:
            #Open output from MPPD and write subtree suspended particles data into sim object
            csv_reader=csv.reader(f)
            for lnum,line in enumerate(csv_reader):
                nhead=1
                if lnum >= nhead:
                    csvgen[lnum-nhead]=int(line[0])
                    csvtimesus=float(line[1])
                    csvsus[lnum-nhead]=float(line[2])
# Assume time given in same order as for exhalation.  But test this.
                    modind=(lnum-nhead)%numfiletimes
                    assert small(csvtime[modind]-csvtimesus), "Exh,Sus file time problem"
                    numfiledatalines_sus=lnum-nhead+1
#           assert numfiledatalines%numfiletimes==0, "Exh, Dep file numgen problem"
            assert numfiledatalines_sus==numfiledatalines_dep, "Sus, Dep files diff length"
            numgen=round(numfiledatalines_sus / numfiletimes)
            # The data items we want to write are b+k*numfiletimes+m
            # for k=0... numgen-1 and m=1... ntimes 
            for k in range(numgen):
                for m in range(1,ntimes+1):
                    dataitem=b+k*numfiletimes+m
                    sim.mass1Dsuspended[idist,csvgen[dataitem],m+a]= csvsus[dataitem]

def samesubset(val,nvals,timea,timeb):
    """Find positions a,b in arrays timea,timeb that equal val
    Require that nvals consective values agree to within TOL"""

    a=(np.abs(timea-val)).argmin()
    b=(np.abs(timeb-val)).argmin()
    suba=timea[a:a+nvals]
    subb=timeb[b:b+nvals]
    assert small(suba[0]-val), "expected time not found in array"
    assert small(suba-subb), "time arrays dont match up"
    return a,b

def call1Donly(sim: simclass.SimClass):
    """Call 1D code with copied (not computed) 3D outflows.  Only valid if flows between 3D and 1D
    compartments are nearly the same from breath to breath"""

    tstart1D=sim.inhales[sim.currbreath][0]
    tend1D=sim.exhales[sim.currbreath][1]
    tendinh=sim.inhales[sim.currbreath][1]

    sim.multiTimeCompleted=tend1D
    sim.multiTimePartial=tend1D
    newtimes=round((tend1D-tstart1D)/sim.dtio)
    inhtimes=round((tendinh-tstart1D)/sim.dtio)
    a=np.abs(tstart1D+sim.dtio-sim.time).argmin()
    assert small(np.abs(tstart1D+sim.dtio-sim.time[a])), "sim not ready for new times"
    writedir=f'multiResults/{sim.currbreath}_inh'
    if os.system(f'mkdir -p {writedir}'):
        sys.exit(f'call1Donly: Cannot create {writedir}')
    for ind in range(sim.ndist):
        writefile=f'{writedir}/airvol_pmass_{name1D[ind]}.csv'
# Flow integrated volume. Volumes must be maxxed with zero, since
# MPPD cannot tolerate the tiniest negative flow volume.
        write_csv_3D1D(writefile,sim.time[a-1:a+newtimes]-sim.time[a-1],
                np.maximum(0.,
                sim.specairvols[ind,a-1:a+newtimes]-sim.specairvols[ind,a-1]),
                sim.mass3Dto1D[ind,a-1:a+newtimes]-sim.mass3Dto1D[ind,a-1])

# Copy constant values of 1D fields during coming inhale phase
    for ind in range(sim.ndist):
        sim.mass1Dto3D[ind,a:a+inhtimes]=sim.mass1Dto3D[ind,a-1]

# Call 1D MPPD code since need mass1Dto3D values for the following.
    call1D(sim)

# Dump sim object to file.  Put in a time-named directory with _exh suffix.
    writedir=f'multiResults/{sim.currbreath}_exh'
    if os.system(f'mkdir -p {writedir}'):
        sys.exit(f'call1Donly: Cannot create {writedir}')
    file_simdump=f'{writedir}/sim.yaml'
    with open(file_simdump,'w') as f:
        yaml.dump(sim,f)


