"""comm3D.py--this module generates input files read by the 3D CFPD code"""

import sys
import os
import numpy as np
import simclass
from data import name3D,name1D,injectors,shortinjectors

TOL=1.e-8
small=lambda arr : np.amax(np.abs(arr)) < TOL

def genparticle_dmdt_inlet(sim: simclass.SimClass):
    """FOAMSPECIFIC ... writes input file giving dm/dt of injected particles"""

    SOI=sim.inhales[sim.currbreath][0]   #SOI start time of current inhalation injection
    #using SOI and particle_dmdt_inlet_profile.txt, generate particle_dmdt_inlet.txt which is dm/dt
    #for current particle injection relative to SOI time
    absolute_inlet_file='particle_dmdt_inlet_profile.txt'
    try:
        with open(absolute_inlet_file,'r') as absfile:
            abscontents=absfile.readlines()
            abstime=[]
            absdmdt=[]
            for line in abscontents:
                splline=line.replace('(',' ').replace(')',' ').split()
                if len(splline)>0:
                    abstime.append(float(splline[0]))
                    absdmdt.append(float(splline[1]))
    except:
        sys.exit(f'{absolute_inlet_file} does not exist\n')
        
    reltime=[]
    for i,_ in enumerate(abstime):
        reltime.append(abstime[i]-SOI) # times are computed relative to SOI
    relative_inlet_file='particle_dmdt_inlet.txt' # name of file read by 3D code on current breath
    with open(relative_inlet_file,'w') as relfile:
        relfile.write(f'(\n')
        for i,_ in enumerate(reltime): # only write non-negative relative times
            if reltime[i]>=0:
                relfile.write(f'( {str(reltime[i])}\t\t{absdmdt[i]} ) \n')
            else:
                if reltime[i+1]>0:   # in this case a fraction of dm/dt occurs at positive relative time
                    intrp=(reltime[i+1]*absdmdt[i]-
                            reltime[i]*absdmdt[i+1])/(reltime[i+1]-reltime[i])
                    relfile.write(f'( 0.\t\t{intrp} ) \n')

        #extra dtio of zero injection to prevent EOF error
        relfile.write(f'( {str(reltime[-1]+sim.dtio)}\t\t 0.0 ) \n') 
        relfile.write(f')\n')

def editentry(entryname,entryvalue,filename):
    """Scans for entryname in filename and adds entryvalue"""

    #Caution:  will replace multiple occurrences, without warning.
    #We are only expecting to replace one occurrence.
    if entryname=="#include":
        char=" "
    else:
        char=";"
#Note:  .replace('^',''), since carat assumed to mean beginning of line.
    entryname1=entryname.replace('^','')
    com=f"sed -i '/{entryname}/ " \
            f"c\{entryname1}\t\t{entryvalue}{char}' {filename}"
    if (os.system(com)):
        sys.exit(f'>>> error editing {entryname} in {filename}')

def readinjection(sim: simclass.SimClass):
    """FOAMSPECIFIC-- read injection history from 3D output file"""

#Regular expression processing of text output file necessitated since injection
#data cannot be simply obtained from postProcessing directory.
    import re
    #Read whole file into array
    try:
        with open(sim.output3Dfile,'r') as f:
            fcontents=f.readlines()
    except:
        sys.exit(f'{sim.output3Dfile} cannot be read.\n')

    ntimes=0
    for line in fcontents:
        if re.search('^Time = ',line):
            ntimes=ntimes+1

    #Parse data out and insert in injtime[], mass[inj#][] arrays
    #Assumed formats in Lung.out OpenFOAM log file
    #   Time = 3.00003e-07                 <Time>=float(line.split()[2])
    #   Injector inletInjection_inlet:     <injector name>=line.split()[1][:-1]
    #   - mass introduced         = 0      <mass introduced>=float(line.split()[4])
    times=np.zeros(ntimes)
    numinj=len(injectors)
    massinj=np.zeros(shape=(numinj,ntimes))
    currtimeind=-1
    newtime=False
    for i,line in enumerate(fcontents):
        if re.search('^Time = ',line):
            try:
                newtime=True
                timecandidate=float(line.split()[2])
            except:
                sys.exit('Could not convert to float Time.')
        elif re.search('mass introduced',line):
            if newtime:
                newtime=False
                currtimeind+=1
                times[currtimeind]=timecandidate
            try:
                ind=injectors.index(fcontents[i-2].split()[1].replace(':',''))
                massinj[ ind, currtimeind ]=float(line.split()[4])
            except:
                sys.exit('Wrong injector name or bad mass')
    #Pass 3:  Wipe out blocks of time that are followed by earlier time blocks.  (It is assumed that 
    #the earlier time blocks are made obsolete by the superceding blocks)
    numtimes=currtimeind+1
    
    if numtimes==0:
        print(f'readinjection warning: {sim.output3Dfile} contains no injection times')
        print(f'readinjection warning:  returning zeros for sim.massInhaled',flush=True)
        sim.massInhaled[:]=0.
        return

    valid=np.zeros(numtimes,dtype=bool)
    valid[:]=True
    currtime=times[numtimes-1]
    valid[numtimes-1]=True
    for i in range(numtimes-2,-1,-1): #This steps back from numtimes-2 to 0 inclusive
        if times[i]>=currtime:
            valid[i]=False
        else:
            currtime=times[i]

    nextind=0
    for i in range(numtimes):
        if valid[i]:
            times[nextind]=times[i]
            massinj[:,nextind]=massinj[:,i]
            nextind+=1
    numtimes=nextind
    times=times[:numtimes]
    massinj=massinj[:,:numtimes]
# massinj contains some zero values in injection time interval that just mean the injector did
# not inject.  These values need to be set to the injection amount at the last injection time when
# the injector was active.
    for i in range(1,numtimes):
        for j in range(numinj):
            if massinj[j,i] < massinj[j,i-1]:
                if massinj[j,i] != 0.:
                    sys.exit(f'Supposedly decreased injection at time {times[i]}\n')
                else:
                    massinj[j,i] = massinj[j,i-1]

    #Interpolate mass flows for injectors to standard io times.
    #Only last injector len(injectors)-1 (inlet) data is used.
    #Injector < len(injectors)-1 data (outlet) is discarded but could be used in future.
    sim.massInhaled=np.interp(sim.time,times,massinj[len(injectors)-1,:])

def genparticle_dmdt_outlets(sim: simclass.SimClass):
    """FOAMSPECIFIC-- write injection files for particles returning from subtrees"""
    
    SOI=sim.exhales[sim.currbreath][0]
    tend3D=sim.exhales[sim.currbreath][1]
    noff=indexwithtol(SOI,sim.time)
    nend=indexwithtol(tend3D,sim.time)
    abstime=sim.time[noff:nend+1]
    reltime=[]
    for i,_ in enumerate(abstime):
        reltime.append(abstime[i]-SOI)
    for idist in range(sim.ndist):
        exh=sim.mass1Dto3D[idist,noff:nend+1]
        absdmdt=np.zeros(nend+1-noff)
        for i in range(1,nend+1-noff):
            absdmdt[i]=exh[i]-exh[i-1]
        relative_outlet_file=f'particle_dmdt_{name3D[idist]}.txt'
        with open(relative_outlet_file,'w') as relfile:
            relfile.write(f'(\n')
            for i,_ in enumerate(reltime):
                relfile.write(f'( {str(reltime[i])}\t\t{absdmdt[i]} ) \n')
            relfile.write(f'( {str(reltime[-1]+sim.dtio)}\t\t 0.0 ) \n')
            relfile.write(f')\n')

def indexwithtol(val,arr):
    """Find index that val is in array arr within tolerance TOL"""

    a=(np.abs(arr-val)).argmin()
    assert small(arr[a]-val), "cant find time in array"
    return a

