"""simclass.py--- class and methods for 'sim' object which holds all necessary 3D and 1D information
of the computation
"""

import numpy as np
import sys
import os
from data import name3D,tstart,tend,dtio,inhales,exhales,ndist,ngen, \
        params1D,params3D,output3Dfile,inhaleMasses,depPatch, \
        nDepPatch,particleDiameter,particleDensity,reuse3D

TOL=1.e-8
small=lambda arr : np.amax(np.abs(arr)) < TOL

class SimClass:
    def __init__(self):
        """Create sim object that holds simulation data.  Object is either created from scratch at
        beginning of simulation, or is created using previous simulation data stored in 
        sim.yaml file on disk.
        """

        self.multiTimeCompleted=tstart
        self.multiTimePartial=tstart
        self.output3Dfile=output3Dfile
# Determine texhend which is time of exhale end in breath
# where 'tend' is situated.  We allocate for sim object out to texhend.
        if tend+TOL<inhales[0][0] or tend-TOL>exhales[-1][1]:
            sys.exit('Illegal tend value')
        for i in range(len(exhales)):
            if tend-TOL<=exhales[i][1]:
                texhend=exhales[i][1]
                break

# Determine current breath number.  
        if tstart+TOL<inhales[0][0] or tstart-TOL>exhales[-1][1]:
            sys.exit('Illegal tstart value')
        for i in range(len(exhales)):
            if tstart-TOL<=exhales[i][1]:
                self.currbreath=i
                break

# In case of reuse3D=True, we require that tstart exactly equals the end of the current breath
# and we require tend exactly equals the end of a future breath.
        if reuse3D:
            if np.abs(tstart - exhales[self.currbreath][1]) > TOL:
                sys.exit('Illegal tstart value')
            tendmatch=False
            for i in range(self.currbreath+1,len(exhales)):
                if np.abs(tend - exhales[i][1]) < TOL:
                    tendmatch=True
            if not tendmatch:
                sys.exit('Illegal tend value')

#Define an orientation array that says flow is either into (+1) or out of (-1)
#1D code. the index corresponds to time interval = [index*dtio,index*dtio+dtio]
        norient=int(round(exhales[-1][1]/dtio))
        floworient=np.zeros(norient)
        torient=inhales[0][0]
        tind=0
        floworient[tind]=1
        tind+=1
        torient+=dtio
        for ind,_ in enumerate(exhales):
            while round((torient-inhales[ind][1])/dtio)<0:
                floworient[tind]=1
                tind+=1
                torient+=dtio
            while round((torient-exhales[ind][1])/dtio)<0:
                floworient[tind]=-1
                tind+=1
                torient+=dtio

# Determine if a historic sim file needs to be read based on tstart and determine
# its name which is a function of breath number.
        if tstart-TOL<=inhales[0][1]:
            historicsimfile=None
            nio=int(round(texhend/dtio)) #if tstart > 0, get zero pad in sim.
            self.time=np.zeros(1+nio)
            for i,_ in enumerate(self.time):
                self.time[i]=i*dtio
            self.massInhaled=np.zeros(1+nio) # Obtain by parsing Lung.out
            self.mass3Ddeposited=np.zeros(shape=(nDepPatch,1+nio))
            self.mass3Dsuspended=np.zeros(1+nio)
            self.specflowrate=np.zeros(shape=(ndist,1+nio))
            self.specairvols=np.zeros(shape=(ndist,1+nio))
            self.recflowrate3D=np.zeros(shape=(ndist,1+nio))
            self.mass3Dto1D=np.zeros(shape=(ndist,1+nio))
            self.mass1Dto3D=np.zeros(shape=(ndist,1+nio))
#The i'th DISTal outlet is given by ordering in post3DInhale.py
#The i'th GENeration is numbered starting from proximal (1D) going to distal
            self.mass1Ddeposited=np.zeros(shape=(ndist,ngen,1+nio)) 
            self.mass1Dsuspended=np.zeros(shape=(ndist,ngen,1+nio)) 
            self.massExhaled=np.zeros(1+nio)

        else:
            if reuse3D:
                historicsimfile=f'multiResults/{self.currbreath}_exh/sim.yaml'
            elif tstart-TOL<=inhales[self.currbreath][1]:
                historicsimfile=f'multiResults/{self.currbreath-1}_exh/sim.yaml'
            else:
                historicsimfile=f'multiResults/{self.currbreath}_inh/sim.yaml'

            import yaml
            try:
                with open(historicsimfile,'r') as f:
                    historicsim=yaml.unsafe_load(f)
            except:
                sys.exit(f'{historicsimfile} had problem loading\n')

            self.time=historicsim.time
            self.massInhaled=historicsim.massInhaled
            self.mass3Ddeposited=historicsim.mass3Ddeposited
            self.mass3Dsuspended=historicsim.mass3Dsuspended
            self.specflowrate=historicsim.specflowrate
            self.specairvols=historicsim.specairvols
            self.recflowrate3D=historicsim.recflowrate3D
            self.mass3Dto1D=historicsim.mass3Dto1D
            self.mass1Dto3D=historicsim.mass1Dto3D
            self.mass1Ddeposited=historicsim.mass1Ddeposited
            self.mass1Dsuspended=historicsim.mass1Dsuspended
            self.massExhaled=historicsim.massExhaled

            nio=int(round((texhend-self.time[-1])/dtio))
            if nio>0:
                newtimes=self.time[-1]+dtio*np.array(range(1,nio+1))
                self.time=np.append(self.time,newtimes,axis=0)
                self.massInhaled=np.append(self.massInhaled,np.zeros(nio),axis=0)
                self.mass3Ddeposited=np.append(self.mass3Ddeposited,np.zeros(shape=(nDepPatch,nio)),axis=1)
                self.mass3Dsuspended=np.append(self.mass3Dsuspended,np.zeros(nio),axis=0)
                self.specflowrate=np.append(self.specflowrate,np.zeros(shape=(ndist,nio)),axis=1)
                self.specairvols=np.append(self.specairvols,np.zeros(shape=(ndist,nio)),axis=1)
                self.recflowrate3D=np.append(self.recflowrate3D,np.zeros(shape=(ndist,nio)),axis=1)
                self.mass3Dto1D=np.append(self.mass3Dto1D,np.zeros(shape=(ndist,nio)),axis=1)
                self.mass1Dto3D=np.append(self.mass1Dto3D,np.zeros(shape=(ndist,nio)),axis=1)
                self.mass1Ddeposited=np.append(self.mass1Ddeposited,np.zeros(shape=(ndist,ngen,nio)),axis=2)
                self.mass1Dsuspended=np.append(self.mass1Dsuspended,np.zeros(shape=(ndist,ngen,nio)),axis=2)
                self.massExhaled=np.append(self.massExhaled,np.zeros(nio),axis=0)

        self.loadedhistoricsimfile=historicsimfile
        self.tstart=tstart
        self.tend=tend
        self.dtio=dtio
        self.inhales=inhales
        self.exhales=exhales
        
        self.ndist=ndist
        self.ngen=ngen
        self.nDepPatch=nDepPatch
        self.params1D=params1D
        self.params3D=params3D
        self.particleDiameter=particleDiameter
        self.particleDensity=particleDensity

        timeCurrInhale=inhales[self.currbreath][1]-inhales[self.currbreath][0]
        self.inhaleMasses=inhaleMasses 
        currInhaleMass=inhaleMasses[self.currbreath]
        self.parcelsPerKilo=params3D["inhaleParcelsPerSecond"]*timeCurrInhale/currInhaleMass

        if not reuse3D:
#for all outlets read specified flows from include files accessed by processor0/{self.tstart}/U.air
            self.read3Dflowprofile() 
# Integrate profile to give air volumes for each outlet.
            for i,_ in enumerate(name3D):
                for j,_ in enumerate(self.time):
                    if j==0:
                        self.specairvols[i,j]=0.
                    else:    
                        avgflow=0.5*(self.specflowrate[i,j]+self.specflowrate[i,j-1])
                        timeincr=self.time[j]-self.time[j-1]
                        self.specairvols[i,j]=self.specairvols[i,j-1] + timeincr*avgflow*floworient[j-1]

    def read3Dflowprofile(self):
        """FOAMSPECIFIC-- read specified flows for each 3D/1D outlet"""

        import re
        strtime=str(round(self.tstart, 6 ))
        strtime=strtime.rstrip('0').rstrip('.')
        Uairfile='processor0/'+strtime+'/U.air'
        try:
            with open(Uairfile,'r') as f:
                fcontents=f.readlines()
        except:
            sys.exit(f'read3Dflowprofile:  Trouble reading {Uairfile}\n')
        for i,_ in enumerate(fcontents):
            if re.match(r'^\s*boundaryField\s*$',fcontents[i]):
                fcontents=fcontents[i:]
                break
        for i,outlet in enumerate(name3D):
            flowfile=''
            restr=r'^\s*'+outlet+r'\s*$'
            for j,_ in enumerate(fcontents):
                if re.match(restr,fcontents[j]):  #outlet clause found
                    for line in fcontents[j+1:]:
                        if line.split()[0]=='file':  #file entry in outlet clause found
                            flowfile=line.split()[1].replace("'","")
                            flowfile=flowfile.replace('"','')
                            flowfile=flowfile.replace(';','')
                            break
                        elif re.search(r'}',line): #clause parsed w/o finding file entry
                            sys.exit(f'read3Dflowprofile: {outlet} clause found'
                                ' but no file entry in U.air found\n')
                    if flowfile=='':  #EOF reached without finding file entry
                        sys.exit(f'read3Dflowprofile: {outlet} clause found but'
                              ' no file entry found and EOF in U.air reached\n')
                    break
            if flowfile=='':
                sys.exit(f'read3Dflowprofile: {outlet} clause not found\n')

            with open(flowfile,'r') as g:
                gcontents=g.readlines()
                time=[]
                specflow=[]
                for line in gcontents:
                    splline=line.replace('(',' ').replace(')',' ').split()
                    if len(splline)>0:
                        time.append(float(splline[0]))
                        specflow.append(float(splline[1]))
            self.specflowrate[i,:]=np.interp(self.time,time,specflow)

    def clone3Ddata(self,clonedbreath):
        srca=indexwithtol(self.inhales[clonedbreath][0],self.time)+1
        srcb=indexwithtol(self.exhales[clonedbreath][1],self.time)+1
        desta=indexwithtol(self.inhales[self.currbreath][0],self.time)+1
        destb=indexwithtol(self.exhales[self.currbreath][1],self.time)+1
# for non-accumulating fields, copy data in [srca:srcb] to [desta:destb]
# for accumulating fields, copy OFFSETTED data in [srca:srcb] to [desta:destb]
# for specairvols, perform integration, since specflowrate may not integrate to zero
#     over a period.
# for recflowrat3D, leave zero to show 3D was not actually run.
        self.massInhaled[desta:destb]=( self.massInhaled[desta-1]-
                self.massInhaled[srca-1] + self.massInhaled[srca:srcb] )
        for ind in range(nDepPatch):
            self.mass3Ddeposited[ind,desta:destb]=( self.mass3Ddeposited[ind,desta-1]-
                    self.mass3Ddeposited[ind,srca-1] + self.mass3Ddeposited[ind,srca:srcb] )
        self.mass3Dsuspended[desta:destb]=self.mass3Dsuspended[srca:srcb]    
        for ind in range(ndist):
            self.specflowrate[ind,desta:destb]=self.specflowrate[ind,srca:srcb]
        for j in range(desta,destb):
            avgflow=0.5*(self.specflowrate[:,j]+self.specflowrate[:,j-1])
            if round((self.time[j]-self.inhales[self.currbreath][1])/dtio)<=0:
                self.specairvols[:,j]=self.specairvols[:,j-1]+dtio*avgflow
            else:    
                self.specairvols[:,j]=self.specairvols[:,j-1]-dtio*avgflow
        for ind in range(ndist):
            self.mass3Dto1D[ind,desta:destb]=( self.mass3Dto1D[ind,desta-1]-
                    self.mass3Dto1D[ind,srca-1] + self.mass3Dto1D[ind,srca:srcb] )
#           self.mass1Dto3D does not get cloned but gets computed by MPPD code.  Although
#           it is recorded, it is ignored since 3D code is not being run.  However, in 
#           principle it can be monitored to check the assumption that 3D injection
#           profile during exhalation is not changing.
        self.massExhaled[desta:destb]=( self.massExhaled[desta-1]-
                self.massExhaled[srca-1] + self.massExhaled[srca:srcb] )

def indexwithtol(val,arr):
    """Find position in arr that is within TOL of val"""

    a=(np.abs(arr-val)).argmin()
    assert small(arr[a]-val), "cant find time in array"
    return a

